{
  description = "Hikarix";

  inputs = {
    nixpkgs.url = "github:nixos/nixpkgs/nixpkgs-unstable";
    flake-utils.url = "github:numtide/flake-utils";
    flake-compat = {
      url = "github:edolstra/flake-compat";
      flake = false;
    };
    gitignore = {
      url = "github:hercules-ci/gitignore.nix";
      inputs.nixpkgs.follows = "nixpkgs";
    };
    zig.url = "github:mitchellh/zig-overlay";
  };

  outputs = _:
    let
      zigVersion = "0.12.0";
    in
    let
      overlays = [
        (final: prev: {
          zig = _.zig.packages.${prev.system}.${zigVersion};
        })
      ];
      systems = builtins.attrNames _.zig.packages;
      inherit (_.gitignore.lib) gitignoreSource;
    in
    _.flake-utils.lib.eachSystem systems (
      system:
      let
        pkgs = import _.nixpkgs { inherit overlays system; };
      in
      {
        packages = {
          default = pkgs.stdenvNoCC.mkDerivation {
            name = "hikarix";
            src = gitignoreSource ./.;
            dontConfigure = true;
            dontInstall = false;
            doCheck = true;
            nativeBuildInputs = [ pkgs.zig ];
            version = "master";
            buildPhase = ''
              mkdir -p .cache
              zig build \
                --cache-dir $(pwd)/zig-cache \
                --global-cache-dir $(pwd)/.cache \
                --prefix $out
            '';
            installPhase = ''
              cp index.html $out
              cp script.js $out
              cp stats.js $out
              cp stats.css $out
              mv $out/bin/main.wasm $out
              rm -rf $out/bin
            '';
            checkPhase = ''
              ls src/*.zig | xargs -I {} zig test {} \
                --cache-dir $(pwd)/zig-cache \
                --global-cache-dir $(pwd)/.cache
            '';
          };
        };
        devShells.default = pkgs.mkShell {
          nativeBuildInputs = with pkgs; [
            entr
            fswatch
            nodePackages_latest.live-server
            process-compose
            zig
          ];
          shellHook =
            let
              mkYaml = name: attrs: (pkgs.formats.yaml { }).generate name attrs;
              processComposeYaml = mkYaml "process-compose.yaml"
                ({
                  processes =
                    let
                      depends_on_zig = {
                        depends_on = {
                          zig = {
                            condition = "process_started";
                          };
                        };
                      };
                    in
                    {
                      zig.command = "echo src/main.zig | entr zig build --verbose";
                      web = {
                        inherit depends_on_zig;
                        command = "live-server";
                      };
                      test = {
                        inherit depends_on_zig;
                        command = "sh -c 'ls src/*.zig | entr -s \"clear && ls src/*.zig | xargs -I {} zig test {}\"'";
                      };
                    };
                });
            in
            ''
              cat ${processComposeYaml} > process-compose.yaml
              zig version
              echo process-compose -f process-compose.yaml
            '';
        };
        formatter = pkgs.nixpkgs-fmt;
      }
    );
}
