var memory = new WebAssembly.Memory({ initial: 100, maximum: 100 });

var importObject = {
  env: {
    memory: memory,
    consoleLog: (ptr, len) => {
      const bytes = new Uint8Array(memory.buffer, ptr, len);
      const message = new TextDecoder("utf-8").decode(bytes);
      console.log(
        "%c" + message,
        "color: #9BBC0F; background-color: #306230; font-family: monospace; font-size: 24px;",
      );
    }
  },
};

WebAssembly.instantiateStreaming(
  fetch("main.wasm"),
  importObject,
).then((result) => {
  const wasmMemoryArray = new Uint8Array(memory.buffer);

  const { draw, init, getBuffer, getBufferHeight, getBufferWidth } =
    result.instance.exports;

  const canvas = document.getElementById("buffer");
  canvas.width = getBufferWidth();
  canvas.height = getBufferHeight();

  const context = canvas.getContext("2d");
  const imageData = context.createImageData(canvas.width, canvas.height);

  var rS = new rStats({
    values: {
      frame: { caption: "Total frame time (ms)", over: 16 },
      raf: { caption: "Time since last rAF (ms)" },
      fps: { caption: "Framerate (FPS)", below: 30 },
    },
  });

  let lastTimestamp = 0;
  const drawCanvas = (timestamp) => {
    const delta = timestamp - lastTimestamp;
    if (delta >= 1000 / 30) {
      lastTimestamp = timestamp;

      rS("frame").start();
      rS("rAF").tick();
      rS("FPS").frame();

      draw(delta);

      const bufferOffset = getBuffer();
      const imageDataArray = wasmMemoryArray.slice(
        bufferOffset,
        bufferOffset + canvas.width * canvas.height * 4,
      );

      imageData.data.set(imageDataArray);
      context.clearRect(0, 0, canvas.width, canvas.height);
      context.putImageData(imageData, 0, 0);

      rS("frame").end();
      rS().update();
    }
    requestAnimationFrame(drawCanvas);
  };

  init();
  requestAnimationFrame(drawCanvas);
});
