const std = @import("std");

// TODO: pub fn angle
pub fn Vec(comptime count: usize, comptime T: type) type {
    return struct {
        data: [count]T,
        const Self = @This();
        const N = count;

        pub usingnamespace switch (count) {
            inline 2 => struct {
                pub inline fn init(_x: T, _y: T) Self {
                    return .{ .data = [_]T{ _x, _y } };
                }
                pub inline fn x(self: Self) T {
                    return self.data[0];
                }
                pub inline fn y(self: Self) T {
                    return self.data[1];
                }
            },
            inline 3 => struct {
                pub inline fn init(_x: T, _y: T, _z: T) Self {
                    return .{ .data = [_]T{ _x, _y, _z } };
                }
                pub inline fn x(self: Self) T {
                    return self.data[0];
                }
                pub inline fn y(self: Self) T {
                    return self.data[1];
                }
                pub inline fn z(self: Self) T {
                    return self.data[2];
                }
            },
            inline 4 => struct {
                pub inline fn init(_x: T, _y: T, _z: T, _w: T) Self {
                    return .{ .data = [_]T{ _x, _y, _z, _w } };
                }
                pub inline fn x(self: Self) T {
                    return self.data[0];
                }
                pub inline fn y(self: Self) T {
                    return self.data[1];
                }
                pub inline fn z(self: Self) T {
                    return self.data[2];
                }
                pub inline fn w(self: Self) T {
                    return self.data[3];
                }
            },
            else => @compileError("Unsupported Vec dimension:" ++ count),
        };

        pub fn from(slice: [N]T) Self {
            return .{ .data = slice };
        }

        pub fn to(self: Self, comptime toType: type) Vec(N, toType) {
            var result = Vec(N, toType){ .data = undefined };
            for (0..count) |i| {
                result.data[i] = switch (T) {
                    u32 => switch (toType) {
                        u32 => self.data[i],
                        f32 => @floatFromInt(self.data[i]),
                        usize => @intCast(self.data[i]),
                        else => @compileError("Unsupported type conversion from " ++ @typeName(T) ++ " to " ++ @typeName(toType)),
                    },
                    usize => switch (toType) {
                        u32 => @intCast(self.data[i]),
                        f32 => @floatFromInt(self.data[i]),
                        usize => self.data[i],
                        else => @compileError("Unsupported type conversion from " ++ @typeName(T) ++ " to " ++ @typeName(toType)),
                    },
                    f32 => switch (toType) {
                        u32 => @intFromFloat(self.data[i]),
                        f32 => self.data[i],
                        usize => @intFromFloat(self.data[i]),
                        else => @compileError("Unsupported type conversion from " ++ @typeName(T) ++ " to " ++ @typeName(toType)),
                    },
                    else => @compileError("Unsupported type conversion from " ++ @typeName(T) ++ " to " ++ @typeName(toType)),
                };
            }
            return result;
        }

        pub fn eq(self: Self, other: Self) bool {
            const epsilon = switch (@typeInfo(T)) {
                .Float => std.math.floatEps(T),
                else => 0,
            };
            return for (self.data, other.data) |a, b| {
                if (@abs(a - b) > epsilon) break false;
            } else true;
        }

        pub fn add(self: Self, other: Self) Self {
            var result = Self{ .data = undefined };
            return for (self.data, other.data, &result.data) |a, b, *res| {
                res.* = a + b;
            } else result;
        }

        pub fn sub(self: Self, other: Self) Self {
            var result = Self{ .data = undefined };
            return for (self.data, other.data, &result.data) |a, b, *res| {
                res.* = a - b;
            } else result;
        }

        pub fn mulScalar(self: Self, scalar: T) Self {
            var result = Self{ .data = undefined };
            return for (self.data, &result.data) |a, *res| {
                res.* = a * scalar;
            } else result;
        }

        pub fn mulVector(self: Self, other: Self) Self {
            var result = Self{ .data = undefined };
            return for (self.data, other.data, &result.data) |a, b, *res| {
                res.* = a * b;
            } else result;
        }

        pub usingnamespace switch (count) {
            inline 2 => struct {
                pub inline fn mulMat(self: Self, mat: Mat(2, 2, T)) Self {
                    const x = self.x() * mat.get(0, 0) + self.y() * mat.get(0, 1);
                    const y = self.x() * mat.get(1, 0) + self.y() * mat.get(1, 1);
                    return Self.init(x, y);
                }
            },
            inline 3 => struct {
                pub inline fn mulMat(self: Self, mat: Mat(3, 3, T)) Self {
                    const x = self.x() * mat.get(0, 0) + self.y() * mat.get(0, 1) + self.z() * mat.get(0, 2);
                    const y = self.x() * mat.get(1, 0) + self.y() * mat.get(1, 1) + self.z() * mat.get(1, 2);
                    const z = self.x() * mat.get(2, 0) + self.y() * mat.get(2, 1) + self.z() * mat.get(2, 2);
                    return Self.init(x, y, z);
                }
            },
            inline 4 => struct {
                pub inline fn mulMat(self: Self, mat: Mat(4, 4, T)) Self {
                    const x = self.x() * mat.get(0, 0) + self.y() * mat.get(0, 1) + self.z() * mat.get(0, 2) + self.w() * mat.get(0, 3);
                    const y = self.x() * mat.get(1, 0) + self.y() * mat.get(1, 1) + self.z() * mat.get(1, 2) + self.w() * mat.get(1, 3);
                    const z = self.x() * mat.get(2, 0) + self.y() * mat.get(2, 1) + self.z() * mat.get(2, 2) + self.w() * mat.get(2, 3);
                    const w = self.x() * mat.get(3, 0) + self.y() * mat.get(3, 1) + self.z() * mat.get(3, 2) + self.w() * mat.get(3, 3);
                    return Self.init(x, y, z, w);
                }
            },
            else => @compileError("Unsupported Vec dimension:" ++ count),
        };

        pub usingnamespace switch (count) {
            inline 2 => struct {
                pub fn rotate(self: Self, angle: T) Self {
                    const cos_theta = std.math.cos(angle);
                    const sin_theta = std.math.sin(angle);
                    const rotation_matrix = Mat(2, 2, T).init(.{
                        .{ cos_theta, -sin_theta },
                        .{ sin_theta, cos_theta },
                    });
                    return self.mulMat(rotation_matrix);
                }
            },
            inline 3 => struct {
                pub fn rotateX(self: Self, angle: T) Self {
                    const cos_alpha = std.math.cos(angle);
                    const sin_alpha = std.math.sin(angle);
                    const rotation_matrix = Mat(3, 3, T).init(.{
                        .{ 1, 0, 0 },
                        .{ 0, cos_alpha, -sin_alpha },
                        .{ 0, sin_alpha, cos_alpha },
                    });
                    return self.mulMat(rotation_matrix);
                }
                pub fn rotateY(self: Self, angle: T) Self {
                    const cos_beta = std.math.cos(angle);
                    const sin_beta = std.math.sin(angle);
                    const rotation_matrix = Mat(3, 3, T).init(.{
                        .{ cos_beta, 0, sin_beta },
                        .{ 0, 1, 0 },
                        .{ -sin_beta, 0, cos_beta },
                    });
                    return self.mulMat(rotation_matrix);
                }
                pub fn rotateZ(self: Self, angle: T) Self {
                    const cos_gamma = std.math.cos(angle);
                    const sin_gamma = std.math.sin(angle);
                    const rotation_matrix = Mat(3, 3, T).init(.{
                        .{ cos_gamma, -sin_gamma, 0 },
                        .{ sin_gamma, cos_gamma, 0 },
                        .{ 0, 0, 1 },
                    });
                    return self.mulMat(rotation_matrix);
                }
            },
            inline 4 => struct {},
            else => @compileError("Unsupported Vec dimension:" ++ count),
        };

        pub fn len(self: Self) T {
            var result: T = 0;
            return std.math.sqrt(for (self.data) |a| {
                result += a * a;
            } else result);
        }

        pub usingnamespace switch (count) {
            inline 2 => struct {
                pub fn angle(self: Self) T {
                    return std.math.atan2(self.y(), self.x());
                }
            },
            inline 3 => struct {},
            inline 4 => struct {},
            else => @compileError("Unsupported Vec dimension:" ++ count),
        };

        pub fn normalize(self: Self) Self {
            const l = self.len();
            var result = Self{ .data = undefined };
            return for (self.data, &result.data) |a, *res| {
                res.* = a / l;
            } else result;
        }

        pub fn clamp(self: Self, min: Self, max: Self) Self {
            var result = Self{ .data = undefined };
            return for (self.data, min.data, max.data, &result.data) |a, mmin, mmax, *res| {
                res.* = switch (@typeInfo(T)) {
                    .Float => @ceil(@min(@max(a, mmin), mmax)),
                    else => @min(@max(a, mmin), mmax),
                };
            } else result;
        }
    };
}

pub fn Mat(comptime cols: usize, comptime rows: usize, comptime T: type) type {
    return struct {
        data: [cols * rows]T,

        Cols: usize = cols,
        Rows: usize = rows,

        const Self = @This();

        // TODO: use something like @memset
        pub fn init(data: [cols][rows]T) Self {
            var result = Self{ .data = undefined };
            return for (0..cols) |c| {
                for (0..rows) |r| {
                    result.data[c * rows + r] = data[c][r];
                }
            } else result;
        }

        pub fn initAsIdentity() Self {
            var result = Self{ .data = undefined };
            return for (0..cols) |c| {
                for (0..rows) |r| {
                    result.data[c * rows + r] = if (c == r) 1 else 0;
                }
            } else result;
        }

        pub fn get(self: Self, col: usize, row: usize) T {
            return self.data[col * rows + row];
        }
    };
}

pub fn Ray(comptime count: usize, comptime T: type) type {
    return struct {
        origin: Vec(count, T),
        direction: Vec(count, T),
        distance: T,

        const Self = @This();

        pub fn init(origin: Vec(count, T), direction: Vec(count, T), distance: T) Self {
            return .{ .origin = origin, .direction = direction, .distance = distance };
        }
    };
}

// Tests //

const types = [_]type{ usize, u8, u16, u32, u64, f16, f32, f64 };
const dimensions = [_]usize{ 2, 3, 4 };

// Vec

test "Vec x, y, z, w components" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v = Vec(n, T).from([_]T{1} ** n);
            switch (n) {
                2 => {
                    try std.testing.expectEqual(v.x(), 1);
                    try std.testing.expectEqual(v.y(), 1);
                },
                3 => {
                    try std.testing.expectEqual(v.x(), 1);
                    try std.testing.expectEqual(v.y(), 1);
                    try std.testing.expectEqual(v.z(), 1);
                },
                4 => {
                    try std.testing.expectEqual(v.x(), 1);
                    try std.testing.expectEqual(v.y(), 1);
                    try std.testing.expectEqual(v.z(), 1);
                    try std.testing.expectEqual(v.w(), 1);
                },
                else => unreachable,
            }
        }
    }
}

test "Vec initFrom" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(std.mem.eql(T, &v.data, &[_]T{1} ** n));
        }
    }
}

test "Vec to" {
    inline for ([_]type{ usize, u32, f32 }) |fromType| {
        inline for ([_]type{ usize, u32, f32 }) |toType| {
            inline for (dimensions) |n| {
                const v = Vec(n, fromType).from([_]fromType{1} ** n).to(toType);
                try std.testing.expect(std.mem.eql(toType, &v.data, &[_]toType{1} ** n));
            }
        }
    }
}

test "Vec eq" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const epsilon = switch (@typeInfo(T)) {
                .Float => std.math.floatEps(T),
                else => 0,
            };
            const v = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(v.eq(v.mulScalar(1 + epsilon)));
        }
    }
}

test "Vec add" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v1 = Vec(n, T).from([_]T{1} ** n);
            const v2 = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(std.mem.eql(T, &v1.add(v2).data, &[_]T{2} ** n));
        }
    }
}

test "Vec sub" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v1 = Vec(n, T).from([_]T{1} ** n);
            const v2 = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(std.mem.eql(T, &v1.sub(v2).data, &[_]T{0} ** n));
        }
    }
}

test "Vec mulScalar" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(std.mem.eql(T, &v.mulScalar(2).data, &[_]T{2} ** n));
        }
    }
}

test "Vec mulVector" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v1 = Vec(n, T).from([_]T{1} ** n);
            const v2 = Vec(n, T).from([_]T{1} ** n);
            try std.testing.expect(std.mem.eql(T, &v1.mulVector(v2).data, &[_]T{1} ** n));
        }
    }
}

test "Vec mulMat" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const v = Vec(n, T).from([_]T{1} ** n);
            const m = Mat(n, n, T).init(.{.{1} ** n} ** n);
            try std.testing.expect(std.mem.eql(T, &v.mulMat(m).data, &[_]T{n} ** n));
        }
    }
}

test "Vec rotate" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            if (@typeInfo(T) != .Float) continue;
            if (n == 2) {
                const v = Vec(n, T).from([_]T{ 1, 1 });
                const angle = std.math.pi / 2.0;
                const rotated = v.rotate(angle);
                const expected = Vec(n, T).from([_]T{ -1, 1 });
                try std.testing.expect(rotated.eq(expected));
            }
        }
    }
}

test "Vec length" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            switch (n) {
                2 => {
                    const v = Vec(n, T).from([_]T{ 1, 1 });
                    switch (@typeInfo(T)) {
                        .Int => try std.testing.expectEqual(@as(T, std.math.sqrt(2)), v.len()),
                        .Float => try std.testing.expectEqual(@as(T, std.math.sqrt(2.0)), v.len()),
                        else => unreachable,
                    }
                },
                3 => {
                    const v = Vec(n, T).from([_]T{ 1, 1, 1 });
                    switch (@typeInfo(T)) {
                        .Int => try std.testing.expectEqual(@as(T, std.math.sqrt(3)), v.len()),
                        .Float => try std.testing.expectEqual(@as(T, std.math.sqrt(3.0)), v.len()),
                        else => unreachable,
                    }
                },
                4 => {
                    const v = Vec(n, T).from([_]T{ 1, 1, 1, 1 });
                    switch (@typeInfo(T)) {
                        .Int => try std.testing.expectEqual(@as(T, std.math.sqrt(4)), v.len()),
                        .Float => try std.testing.expectEqual(@as(T, std.math.sqrt(4.0)), v.len()),
                        else => unreachable,
                    }
                },
                else => unreachable,
            }
        }
    }
}

test "Vec angle" {
    inline for (types) |T| {
        if (@typeInfo(T) != .Float) continue;
        if (T == f16) continue; // atan2 not implemented for f16 in std.math
        inline for (dimensions) |n| {
            if (n != 2) continue; // TODO
            const v = switch (n) {
                2 => Vec(n, T).from([_]T{ 1, 1 }),
                else => unreachable,
            };
            try std.testing.expectApproxEqAbs(v.angle(), std.math.pi / 4.0, std.math.floatEps(T));
        }
    }
}

test "Vec normalize" {
    inline for (types) |T| {
        if (@typeInfo(T) != .Float) continue;
        inline for (dimensions) |n| {
            const v = switch (n) {
                2 => Vec(n, T).from([_]T{ 1, 2 }),
                3 => Vec(n, T).from([_]T{ 1, 2, 3 }),
                4 => Vec(n, T).from([_]T{ 1, 2, 3, 4 }),
                else => unreachable,
            };
            try std.testing.expectApproxEqAbs(v.normalize().len(), 1.0, std.math.floatEps(T));
        }
    }
}

test "Vec clamp" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const clampLow = Vec(n, T).from([_]T{1} ** n);
            const clampHigh = Vec(n, T).from([_]T{9} ** n);

            const vecLow = Vec(n, T).from([_]T{0} ** n);
            const vecHigh = Vec(n, T).from([_]T{10} ** n);

            try std.testing.expect(vecLow.clamp(clampLow, clampHigh).eq(clampLow));
            try std.testing.expect(vecHigh.clamp(clampLow, clampHigh).eq(clampHigh));
        }
    }
}

// Mat

test "Mat init" {
    inline for (types) |T| {
        inline for (dimensions) |rows| {
            inline for (dimensions) |cols| {
                const m = Mat(cols, rows, T).init(.{.{1} ** rows} ** cols);
                try std.testing.expect(std.mem.eql(T, &m.data, &[_]T{1} ** rows ** cols));
            }
        }
    }
}

test "Mat get" {
    inline for (types) |T| {
        inline for (dimensions) |rows| {
            inline for (dimensions) |cols| {
                const m = Mat(cols, rows, T).init(.{.{1} ** rows} ** cols);
                try std.testing.expectEqual(m.get(0, 0), 1);
            }
        }
    }
}

// Ray

test "Ray init" {
    inline for (types) |T| {
        inline for (dimensions) |n| {
            const start = Vec(n, T).from([_]T{1} ** n);
            const end = Vec(n, T).from([_]T{1} ** n);
            const r = Ray(n, T).init(start, end, 0);
            try std.testing.expect(r.origin.eq(Vec(n, T).from([_]T{1} ** n)));
            try std.testing.expect(r.direction.eq(Vec(n, T).from([_]T{1} ** n)));
        }
    }
}
